__all__ = [
    '__project__', '__title__', '__author__', '__copyright__', '__license__',
    '__version__', '__mail__', '__maintainer__', '__status__'
]

__project__ = 'PINA'
__title__ = "pina"
__author__ = "PINA Contributors"
__copyright__ = "Copyright 2021-2023, PINA Contributors"
__license__ = "MIT"
__version__ = "0.1"
__mail__ = 'demo.nicola@gmail.com, dario.coscia@sissa.it'  # TODO
__maintainer__ = __author__
__status__ = "Alpha"
__packagename__ = "pina-mathlab"
