__all__ = [
    'PINN',
    'GAROM',
    'SupervisedSolver',
    'SolverInterface'

]

from .garom import GAROM
from .pinn import PINN
from .supervised import SupervisedSolver
from .solver import SolverInterface
